/*
 * Copyright (c) 2001-2003 The Trustees of Indiana University.  
 *                         All rights reserved.
 * Copyright (c) 1998-2001 University of Notre Dame. 
 *                         All rights reserved.
 * Copyright (c) 1994-1998 The Ohio State University.  
 *                         All rights reserved.
 * 
 * This file is part of the LAM/MPI software package.  For license
 * information, see the LICENSE file in the top level directory of the
 * LAM/MPI source distribution.
 * 
 * $HEADER$
 *
 * $Id: lamnet.c,v 6.12 2003/05/21 19:17:00 jsquyres Exp $
 * 
 *	Function:	- LAM network functions
 */

#include <lam_config.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#ifndef SIOCGIFCONF
#include <sys/sockio.h>
#endif
#include <sys/time.h>
#include <net/if.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <unistd.h>

#include <laminternal.h>
#include <lamnet.h>
#include <portable.h>
#include <terror.h>
#include <typical.h>
#include <lam_network.h>
#include <etc_misc.h>

#define LAMIFMAX	32


/*
 * global functions
 */
int			lamnet_findhosts();
int			lamnet_findorig();
int			lamnet_dups();


/*
 * local functions
 */
static int getifaddr(uint4 *ifaddr, int nifaddr);


/*
 * local variables
 */
static unsigned long	ifbuf[256];


/*
 *	lamnet_findhosts
 *
 *	Function:	- gets the network address of each host
 *			- "localhost" is changed to the local hostname
 *	Accepts:	- node description array
 *			- array size
 *			- bad host index (out)
 */
int
lamnet_findhosts(struct lamnode *lamnet, int nlamnet, int *badhost)
{
  int i;
  char buffer[16];

  if (badhost != NULL) 
    *badhost = -1;
  
  /* Find INET addresses for each node */
  
  for (i = 0; i < nlamnet; ++i) {

    /* Skip empty entries */

    if (lamnet[i].lnd_nodeid == NOTNODEID)
      continue;
    
    /* Get the host network address  */

    if (getinetaddr(lamnet[i].lnd_hname,
		    (unsigned char *) &lamnet[i].lnd_addr.sin_addr) != 0) {
      if (badhost != NULL) 
	*badhost = i;
      
      if (errno == EBADHOST)
	show_help("boot", "resolve-hostname", lamnet[*badhost].lnd_hname,
		  NULL);
      else {
	snprintf(buffer, 16, "%d", errno);
	show_help("boot", "resolve-unknown", buffer, NULL);
      }
      return(LAMERROR);
    }

    /* Nope; it resolved properly, so save it */

    lamnet[i].lnd_addr.sin_family = AF_INET;
  }

  return 0;
}


/*
 *	lamnet_findorig
 *
 *	Function:	- finds the origin host node
 *			- origin is the machine this code is running on
 *			- calls lamnet_findhosts() if necessary
 *			- do matching on the network addresses
 *			- make sure each host is unique (no duplicates)
 *	Accepts:	- network description array
 *			- array size
 *	Returns:	- index of origin or LAMERROR
 */
int
lamnet_findorig(struct lamnode *lamnet, int nlamnet)
{
	int		i, j;
	int		nifaddr;	/* # of interface addresses */
	uint4		ifaddr[LAMIFMAX];

	nifaddr = getifaddr(ifaddr, LAMIFMAX);
	if (nifaddr < 0) return(LAMERROR);

/*
 * Loop over the host entries, comparing their addresses to
 * our local host interface addresses.
 */
	for (i = 0; i < nlamnet; ++i) {

          /* Skip empty entries */

          if (lamnet[i].lnd_nodeid == NOTNODEID)
            continue;

          for (j = 0; j < nifaddr; ++j) {
            if (ifaddr[j] == lamnet[i].lnd_addr.sin_addr.s_addr) {
              return(i);
            }
          }
	}

	errno = EINVAL;
	return(LAMERROR);
}


/*
 *	getifaddr
 *
 *	Function:	- obtains all up interface addresses for local host
 *	Accepts:	- INET address array (returned addresses)
 *			- max array size
 *	Returns:	- number of addresses or LAMERROR
 */
static int
getifaddr(uint4 *ifaddr, int nifaddr)
{
	int		sock;
	int		length;
	int		num;
	int		rem;
	char		*p;
	struct ifconf	config;
	struct ifreq	req, *preq;
	struct sockaddr_in
			*psin;
/*
 * Create an Internet socket.
 */
	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (sock < 0) return(LAMERROR);
/*
 * Get network interface configuration.
 */
	config.ifc_len = sizeof(ifbuf);
	config.ifc_req = (struct ifreq *) ifbuf;

	if (ioctl(sock, SIOCGIFCONF, &config)) {
		close(sock);
		return(LAMERROR);
	}
/*
 * Loop through all interfaces.
 */
	p = (char *) ifbuf;
	rem = config.ifc_len;
	num = 0;

	while ((rem > 0) && (nifaddr > 0)) {
		preq = (struct ifreq *) p;
#if LAM_HAVE_SA_LEN
		length = sizeof(struct sockaddr);

		if (preq->ifr_addr.sa_len > length) {
			length = preq->ifr_addr.sa_len;
		}

		length += sizeof(preq->ifr_name);
#else
		length = sizeof(struct ifreq);
#endif

		rem -= length;
		p += length;

		if (preq->ifr_addr.sa_family != AF_INET) continue;

		memcpy(req.ifr_name, preq->ifr_name, sizeof(req.ifr_name));

		if (ioctl(sock, SIOCGIFFLAGS, &req)) {
			close(sock);
			return(LAMERROR);
		}

		if (req.ifr_flags & IFF_UP) {

			psin = (struct sockaddr_in *) &(preq->ifr_addr);
			*ifaddr = psin->sin_addr.s_addr;
			num++;
			ifaddr++;
			nifaddr--;
		}
	}

	close(sock);
	return(num);
}
